# -*- coding: utf-8 -*-
#############################
# File:    ImgPL.rb         #
# Authors:                  #
# Mattias Roback (matro414) #
# Magnus Suther (magsu191)  #
#############################

# Require the right files, depending on if the language is
# installed to /usr/local/* or not.
if File.exists?("/usr/local/lib/ImgPL") then
  require '/usr/local/lib/ImgPL/rdparse'
  require '/usr/local/lib/ImgPL/tree'
  require '/usr/local/lib/ImgPL/scopehandler'
else
  if not File.exists?("rdparse.rb") then
    abort("Could not find required file 'rdparse.rb'! Aborting.")
  else
    require 'rdparse'
  end
  if not File.exists?("tree.rb") then
    abort("Could not find required file 'tree.rb'! Aborting.")
  else
    require 'tree'
  end
  if not File.exists?("scopehandler.rb") then
    abort("Could not find required file 'scopehandler.rb'! Aborting.")
  else
    require 'scopehandler'
  end
end

begin
  require 'RMagick'
rescue
  abort("Could not find required module 'RMagick'! Aborting.")
end
require 'fileutils'


def printMascot()
	# Prints our mascot: Homer in ascii art!
  if File.exist?("/usr/local/lib/ImgPL/error.txt") then
    puts File.read("/usr/local/lib/ImgPL/error.txt")
  else
    puts File.read("error.txt") if File.exist?("error.txt")
  end
end

##### Modifications to Ruby-classes #####
class Integer
  def split()
    # Split the number, and return an array with each
    # number as separate items. Used to loop through each number, ie
    # for a in 1234
    a = self.to_s.split("").to_a.map{|a|a.to_i}
    a
  end
  def include?(rhs)
    # Check if the value of self contains the number in rhs
    self.to_s.include? rhs.to_s
  end
end

class Fixnum
  def length()
    self
  end
end

class Float
  def split()
    # Split the number, and return an array with each
    # number as separate items. Used to loop through each number, ie
    # for a in 1234
    self.to_s.split(".").to_a.map{|a|a.to_i}
  end
  def length()
    self
  end
  def include?(rhs)
    # Check if the value of self contains the number in rhs
    self.to_s.include? rhs.to_s
  end
end

class Array
  # Expand the Array class with functions to compare
  # the length of self and rhs
  def <(rhs)
    return self.length < rhs.length
  end
  def >(rhs)
    return self.length > rhs.length
  end
  def <=(rhs)
    return self.length <= rhs.length
  end
  def >=(rhs)
    return self.length >= rhs.length
  end
end

class String
  # Expand the String class with functions to compare
  # the length of self and rhs
  def <(rhs)
    return self.length < rhs.length
  end
  def >(rhs)
    return self.length > rhs.length
  end
  def <=(rhs)
    return self.length <= rhs.length
  end
  def >=(rhs)
    return self.length >= rhs.length
  end
end

class Range
  # Expand the Range class with functions to compare
  # the length of self and rhs
  def length()
    self.end-self.begin
  end
  def <(rhs)
    return self.length < rhs.length
  end
  def >(rhs)
    return self.length > rhs.length
  end
  def <=(rhs)
    return self.length <= rhs.length
  end
  def >=(rhs)
    return self.length >= rhs.length
  end
end


##### Tokens #####
class Token
  # Super class for tokens. When tokens are matched
  # with regular expressions, some of then are returned
  # as instances of the subclasses.
  def initialize(value=nil)
    @value = value
  end

  def value
    @value
  end
end

# Define the subclasses of class Token.
# These are all just using the methods in Token, 
# no other functionality is added.
class Indent < Token;end
class Dent < Token;end
class Dedent < Token;end
class Newline < Token;end
class Begin < Token;end
class End < Token;end
class While < Token;end
class For < Token;end
class Do < Token;end
class Break < Token;end
class Print < Token;end
class Get < Token;end
class If < Token;end
class Else < Token;end
class Then < Token;end
class Add_op < Token;end
class Multi_op < Token;end
class Rel_op < Token;end
class Func < Token;end
class Return < Token;end
class Out_op < Token;end
class In_op < Token;end
class Class_token < Token;end
class Not_token < Token;end
class Or_token < Token;end
class And_token < Token;end
class New_token < Token;end
class Power_op < Token;end

class ImgPL
  def initialize
    @parser = Parser.new("ImgPL") do
      @indent_level = 1
      ##### Strings #####
      token(/^\"([^\"])*\"/) {|m|String_type.new(m)}
      ##### Newlines #####
      token(/^(\\n)+/) {|_|Newline.new}
      ##### Indents, Dedents & Dents #####
      token(/^(\\t)+/)do|m|
        # When we find a tab token, we need to count
        # the number of tabs to keep track of the current
        # indentation level.
        # If the indentation increases, returns a Indent.
        # If indentation decreases, returns a Dedent.
        # If the indentation is unchanged, returns a Dent.
        new_indent = m.split("t").length
        dif = (new_indent-@indent_level).abs
        if new_indent > @indent_level
          @indent_level += dif
          Indent.new(dif)
        elsif new_indent == @indent_level
          Dent.new(dif)
        else new_indent
          @indent_level -= dif
          Dedent.new(dif)
        end
      end
      ##### Operators #####
      token(/^[\+|\-]/){|m| Add_op.new(m)}
      token(/^\*\*/) {|m| Power_op.new} # Match square root operator
      token(/^\<\</){|_| Out_op.new}
      token(/^\>\>/){|_| In_op.new}
      token(/^[\*|\/|\%]/){|m| Multi_op.new(m)}
      token(/^((<=)|(>=)|<|>) /){|m| Rel_op.new(m.split(" ").join(" "))}
      ##### Begin & End Tokens #####
      token(/^_begin/){|_| Begin.new}
      token(/^_end/){|_| End.new}
      ##### Keywords and Identifiers #####
      token(/^not\s+in\s+/){|m| Rel_op.new(m.split(" ").join(" "))}
      token(/^is\s+not\s+/){|m| Rel_op.new(m.split(" ").join(" "))}
      token(/^[A-Za-z]\w*/) do |m|
        case m
        when "null" then Null.new
        when "in","is" then Rel_op.new(m)
        when "true","false" then Bool_type.new(m)
        when "or" then Or_token.new(m)
        when "and" then And_token.new(m)
        when "not" then Not_token.new(m)
        when "println","print","errprint" then Print.new(m)
        when "get" then Get.new
        when "if" then If.new
        when "else" then Else.new
        when "then" then Then.new
        when "while" then While.new
        when "for" then For.new
        when "do" then Do.new
        when "break" then Break.new
        when "return" then Return.new
        when "class" then Class_token.new
        when "func" then Func.new
        else
          if m[0..0].scan(/[A-Z]/).empty?
            Identifier.new(m)
          else
            Class_name.new(m)
          end
        end
      end
      ##### Spaces #####
      token(/ +/)
      ##### Int, Float, Range #####
      token(/^\d+\.\.\d+/){|m|s,e = m.split("..");Range_type.new(s.to_i,e.to_i)}
      token(/^\d+\.\d+/) {|m| Float_type.new(m.to_f) }
      token(/^\d+/) {|m| Int_type.new(m.to_i) }
      ##### Class variables #####
      token(/^;[a-z]+\w*/){|m| Class_var.new(m)}
      ##### The rest #####
      token(/^./) {|m| m }
      
      ##### Rules #####
      start :program do
        match(:lines){|line|Program.new(line).eval()}
      end

      rule :lines do
        match(:line,:lines){|line,lines|Lines.new(line,lines)}
        match(:line){|line|Lines.new(line)}
      end
      
      rule :line do
        match(Dent,Newline){|_,_|Line.new()}
        match(Dent,:expr,Newline){|_,expr,_|Line.new(expr)}
        match(Dent,:stmt,Newline){|_,stmt,_|Line.new(nil,stmt)}
      end

      rule :stmt do
        match(:class_def){|stmt|Stmt.new(stmt)}
        match(:func_def){|stmt|Stmt.new(stmt)}
        match(:if_else_stmt){|stmt|Stmt.new(stmt)}
        match(:assign_stmt){|assign|Stmt.new(assign)}
        match(:print_stmt){|print|Stmt.new(print)}
        match(:get_stmt){|get|Stmt.new(get)}
        match(:while_stmt){|stmt|Stmt.new(stmt)}
        match(:for_stmt){|stmt|Stmt.new(stmt)}
        match(Break){|_|Break_stmt.new()}
        match(Return,:expr){|_,expr|Return_stmt.new(expr)}
      end

      rule :class_def do
        match(Class_token,Class_name,Newline,Indent,Begin,Newline,Dent,:class_vars_def,Dent,:class_funcs_def,Newline,Dedent,End){|_,id,_,_,_,_,_,vars,_,funcs,_,_,_|Class_def.new(id,funcs,vars)}
        match(Class_token,Class_name,Newline,Indent,Begin,Newline,Dent,:class_funcs_def,Newline,Dedent,End){|_,id,_,_,_,_,_,block,_,_,_|Class_def.new(id,block)}
      end

      rule :class_call do
        match(Class_name,'.',Identifier,'(',:arg_list,')'){|class_name,_,func_name,_,args,_|Class_call.new(class_name,func_name,args)}
        match(Class_name,'.',Identifier,'(',')'){|class_name,_,func_name,_,_|Class_call.new(class_name,func_name)}
      end

      rule :class_funcs_def do
        match(:class_funcs_def,Newline,Dent,:func_def){|funcs,_,_,func|Class_funcs_def.new(func,funcs)}
        match(:func_def){|func|Class_funcs_def.new(func)}
      end

      rule :class_vars_def do
        match(:class_vars_def,Dent,:class_var_def){|vars,_,var|Class_vars_def.new(var,vars)}
        match(:class_var_def){|var|Class_vars_def.new(var)}
      end

      rule :class_var_def do
        match(Class_var,Newline){|var,_|Class_var_def.new(var)}
      end

      rule :func_def do
        match(Func, Identifier,'(',:param_list,')', Newline, :block){|_,id,_,params,_,_,block| Func_def.new(id, params, block)}
        match(Func, Identifier,'(',')', Newline, :block){|_,id,_,_,_,block| Func_def.new(id, nil, block)}
      end
      
      rule :param_list do
        match(:param_list,',', :param){|params,_,param| Param_list.new(params, param)}
        match(:param){|param| Param_list.new(nil, param)}
      end

      rule :param do
        match(:assign_stmt) {|stmt| Param.new(stmt)}
        match(Identifier) {|id| Param.new(id)}
      end
      
      rule :while_stmt do
        match(While,:condition,Do,Newline,:block){|_,cond,_,_,block|While_stmt.new(cond,block)}
      end

      rule :for_stmt do
        match(For,Identifier,Rel_op,:datatypes,Do,Newline,:block){|_,id,op,data,_,_,block|For_stmt.new(id,op,data,block)}
      end

      rule :if_else_stmt do
        match(:if_stmt,Newline,Dent,:else_if_stmts,Newline,Dent,:else_stmt)do|stmt,_,_,stmt2,_,_,stmt3|
          If_else_stmt.new(stmt,stmt2,stmt3)
        end
        match(:if_stmt,Newline,Dent,:else_if_stmts){|stmt,_,_,stmt2|If_else_stmt.new(stmt,stmt2)}
        match(:if_stmt,Newline,Dent,:else_stmt){|stmt,_,_,stmt2|If_else_stmt.new(stmt,nil,stmt2)}
        match(:if_stmt){|stmt|If_else_stmt.new(stmt)}
      end
      
      rule :if_stmt do
        match(If,:condition,Then,Newline,:block){|_,cond,_,_,block|If_stmt.new(cond,block)}
      end
      
      rule :else_if_stmts do
        match(:else_if_stmts,Newline,Dent,:else_if_stmt){|stmt2,_,_,stmt|Else_if_stmts.new(stmt,stmt2)}
        match(:else_if_stmt){|stmt|Else_if_stmts.new(stmt)}
      end

      rule :else_if_stmt do
        match(Else,If,:condition,Then,Newline,:block){|_,_,cond,_,_,block|Else_if_stmt.new(cond,block)}
      end
      
      rule :else_stmt do
        match(Else,Newline,:block){|_,_,block|Else_stmt.new(block)}
      end

      rule :block do
        match(Indent,Begin,Newline,:lines,Dedent,End){|_,_,_,line,_,_|Block.new(line)}
      end
      
      rule :assign_stmt do
        match(Identifier,'=',:class_call){|id,_,instance|Assign_stmt.new(id,instance)}
        match(Identifier,'=',:expr){|id,_,expr|Assign_stmt.new(id,expr)}
        match(Class_var,'=',:class_call){|id,_,instance|Assign_stmt.new(id,instance)}
        match(Class_var,'=',:expr){|id,_,expr|Assign_stmt.new(id,expr)}
      end

      rule :print_stmt do
        match(Print,Out_op,:expr){|print,_,expr|Print_stmt.new(print.value,expr)}
      end

      rule :get_stmt do
        match(Get,In_op,Identifier){|_,_,id|Get_stmt.new(id)}
      end

      rule :expr do
        match(:add_expr){|add_expr|Expr.new(add_expr)}
        match(:condition){|cond|Expr.new(cond)}
        
      end

      rule :add_expr do
        match(:add_expr,:add_operator,:multi_expr){|add,op,multi|Add_expr.new(add,op,multi)}
        match(:multi_expr){|multi|Add_expr.new(multi)}
      end

      rule :multi_expr do
        match(:multi_expr,:multi_operator,:unary_expr){|multi,op,unary|Multi_expr.new(multi,op,unary)}
        match(:unary_expr){|unary|Multi_expr.new(unary)}
      end

      rule :unary_expr do
        match(:add_operator,:unary_expr){|op,unary|Unary_expr.new(unary,op)}
        match(:power_expr){|power|Unary_expr.new(power)}
      end

      rule :power_expr do
        match(:datatypes){|datatype|Power_expr.new(datatype)}
        match('(',:add_expr,')'){|_,expr,_|Power_expr.new(expr)}
        match(:power_expr,Power_op,:power_expr){|power,_,expr| Power_expr.new(expr,power)}
      end

      rule :datatypes do
        match(Int_type){|int|Datatypes.new(int)}
        match(Float_type){|float|Datatypes.new(float)}
        match(Range_type){|range|Datatypes.new(range)}
        match(Null){|null|Datatypes.new(null)}
        match(:array){|array|Datatypes.new(array)}
        match(String_type){|string|Datatypes.new(string)}
        match(:datatypes,'[',:expr,']'){|expr1,_,expr2,_|Subscript_expr.new(expr1,expr2)}
        match(Class_var,'.',Identifier,'(',:arg_list,')'){|var,_,func,_,args,_|Instance_func_call.new(var,func,args)}
        match(Class_var,'.',Identifier,'(',')'){|var,_,func,_,_|Instance_func_call.new(var,func)}
        match(Identifier,'.',Identifier,'(',:arg_list,')'){|var,_,func,_,args,_|Instance_func_call.new(var,func,args)}
        match(Identifier,'.',Identifier,'(',')'){|var,_,func,_,_|Instance_func_call.new(var,func)}
        match(Identifier,'(',:arg_list,')'){|id,_,args,_|Func_call.new(id,args)}
        match(Identifier,'(',')'){|id,_,_|Func_call.new(id)}
        match(Identifier){|ident|Get_var.new(ident)}
        match(Class_var){|var|Get_var.new(var)}
      end

      rule :array do
        match('[',:elements,']'){|_,elements,_| Array_type.new(elements)}
        match('[',']'){|_,_|Array_type.new()}
      end

      rule :elements do
        match(:elements,',',:expr){|elements,_,expr|Elements.new(expr,elements)}
        match(:expr){|expr|Elements.new(expr)}
      end

      rule :arg_list do
        match(:arg_list,',', :arg){|args,_,arg| Arg_list.new(arg,args)}
        match(:arg){|arg| Arg_list.new(arg)}
      end

      rule :arg do
        match(:expr) {|expr| Arg.new(expr)}
      end

      rule :add_operator do
        match(Add_op){|op|op.value}
      end

      rule :multi_operator do
        match(Multi_op){|op|op.value}
      end

      rule :rel_operator do
        match(Rel_op){|op|op.value}
      end

      rule :condition do
        match(:or_expr)
      end
      
      rule :or_expr do
        match(:or_expr,Or_token,:and_expr){|expr,_,expr2|Or_expr.new(expr,expr2)}
        match(:and_expr){|expr|Or_expr.new(expr)}
      end

      rule :and_expr do
        match(:and_expr,And_token,:not_expr){|expr,_,expr2|And_expr.new(expr,expr2)}
        match(:not_expr){|expr|And_expr.new(expr)}

      end
      rule :not_expr do
        match(Not_token,:not_expr){|token,expr|Not_expr.new(expr,token)}
        match(:rel_expr){|expr|Not_expr.new(expr)}
      end

      rule :rel_expr do
        match(:add_expr,:rel_operator,:add_expr){|expr1,op,expr2|Rel_expr.new(expr1,op,expr2)}
        match('(',:condition,')'){|_,expr,_|Rel_expr.new(expr)}
        match(:datatypes){|expr|Rel_expr.new(expr)}
        match(Bool_type){|bool|Datatypes.new(bool)}
      end
    end
  end

  #################### INTERPRETED MODE ####################
  def run(command = nil)
		begin
			# Make the logger quiet
			@parser.logger.level = Logger::WARN

			if command.nil? then
				loop do
					str = getInput(0)

					if str.eql?("exit") or str.eql?("quit") then
						$stdout.puts "Bye!"
						exit
					end

					# Find import command, and import the given file.
					if not str.scan(/^import\s+\<.+\>/).empty? then
						importFile(str)
						str = ""
					end

					# Add a tab at the beginning of every line.
					# Thanks to this, the token for tab is matched on
					# every line, and we can keep track of the indentation.
					str = str.gsub(/(\\n)/,"\\n\\t").insert(0,"\\t")+"\\n"

					if not str.empty? then
						out = @parser.parse(str)
						# Output the value returned, if any
						$stdout.puts "=> #{out.inspect}" if !out.nil? and !out.is_a?(Null)
					end
					
				end
			else
				# Lets the testcases use this function to pre-format the string
				# before evaluating

				# Add a tab at the beginning of every line.
				# Thanks to this, the token for tab is matched on
				# every line, and we can keep track of the indentation.
				command = command.gsub(/(\\n)/,"\\n\\t").insert(0,"\\t")+"\\n"
				if not command.empty? then
					print "=> "
					# Always return whatever is returned
					@parser.parse(command)
				end
			end
			rescue NoMethodError, Interrupt
				# The user most likely pressed
				# ctrl-d or ctrl-c
				$stdout.puts "Bye!"
				exit
			rescue Parser::ParseError
				printMascot()
				$stderr.puts "<SYNTAX_ERROR>: Failed to interpret input."
				retry
			rescue => e
				printMascot()
				$stderr.puts e
				retry
	  	end
		end


  #################### EVAL FILE ####################
  def eval_file(path, quiet = false)
	begin
		# Make the logger quiet
		@parser.logger.level = Logger::WARN

		file = File.read(path)
		str = file.chomp

		# Add a tab at the beginning of every line.
		# Thanks to this, the token for tab is matched on
		# every line, and we can keep track of the indentation.
		str = str.gsub(/(\n)/,"\\n\\t").insert(0,"\\t")+"\\n\\t\\n"
		str.gsub!(/\t/,"\\t")
		lines = str.split("\\n")
		final_lines = []
		previous_tabs = 1
		code_lines = []

		# Removes empty lines and block-comments.
		count = 0
		lines.each do |line|				
			count += 1 if line.eql?("\\t/*") and count == 0
			code_lines << line if count.eql?(0) and 
				!line.gsub(/\\t+/,"").gsub(/\s*/,"").empty?
			count -= 1 if line.eql?("\\t*/")
		end
		
		# Removes one-line-comments
		lines = []
		raise("<SYNTAX_ERROR>: Block comments do not match.") if count != 0
		code_lines.each do |line|
			index = line.index(/\/\//)
			if index.nil?
				lines << line
			else
				lines << line[0..index-1]
			end
		end
		lines << "\\t"

		#Adds begin and end-tokens before and after a
		#block, in order to keep track of indentation level
		#when several blocks is closed at the same time.
		lines.each do |line|
			tabs = line.scan(/\\t/).length
			if tabs == previous_tabs+1
				final_lines << ("\\t"*tabs)+"_begin "
				final_lines << line.clone
			elsif tabs < previous_tabs
				(previous_tabs-1).downto(tabs)do |i|
					final_lines << ("\\t"*i)+"_end "
				end
				final_lines << line
			else
				final_lines << line
			end
			previous_tabs = tabs.to_i
		end
		str = final_lines.join("\\n")+"\\n"

		out = @parser.parse(str)
		$stdout.puts "Mission complete!" if not quiet
		
	rescue NoMethodError,Interrupt
		# Most likely, the user has pressed ctrl-d or ctrl-c
		# somewhere, lets abort without a crash.
		$stdout.puts "Bye!"
		exit
	rescue SystemStackError
		printMascot()
		$stderr.puts "<STACK_LEVEL_ERROR>: Stack level to deeeeeeeeeeeeeeep."
	rescue Parser::ParseError
		printMascot()
		$stderr.puts "<SYNTAX_ERROR>: Program contains syntactic errors!"
	rescue => e
		printMascot()
		$stderr.puts e
	end
  end
end


def getInput(indent)
	# Let the user type multiline programs in the interpreter.
	# Since we need to keep track of the indentation to be able to
	# insert special characters and tokens, we solve this recursively.
	
  $stdout.print "[ImgPL] " + ("··· " * indent)
  input = gets.chomp

  if input.start_with?("for ","if ","else if ","else ",
                       "while ", "func ", "class ") then
    input += "\\n" + ("\\t"*(indent+1)) + "_begin \\n" + ("\\t"*(indent+1)) + getInput(indent+1)
  end

  # If there is no block, this is an expression that
  # should be returned immediately. But only if it
  # is not already inside a block.
  return input if indent.eql?(0) or input.empty?

  # Get the block. Block is considered complete when 
  # an empty line is input. The code below 
  # is reached only if a block is expected.
  loop do
    $stdout.print "[ImgPL] " + ("··· " * indent)
    line = gets.chomp
    if line.start_with?("for ","if ","else if ","else ",
                        "while ","func ","class ") then
      # Only run if line is not the first line in the block.
      line += "\\n\\t_begin \\n" + ("\\t"*(indent+1)) + getInput(indent+1) + "\\n"
    end
	
    return input + "\\n" + ("\\t"*(indent-1)) + "_end " if line.empty?()
    input += "\\n" + ("\\t"*(indent)) + line
  end
end

def importFile(str)
	# The import command lets users import/evaluate ImgPL files
	# from the interpreter. This is used to make use of functions
	# and/or classes declared in the imported file.
	# Syntax for the import command: import <file.imgpl>
	
	file = str.split(/\s+/)[1].gsub(/[\<|\>]/, "")
	if not file.nil? then
		if not File.extname(file).eql?(".imgpl") then
			file += ".imgpl"
		end
		if File.exists?(file) then
			# Evaluate the file to store classes, functions and variables
			# declared in it.
			eval_file(file, true)
		else
			# File not found, raise error
			raise("Import failed, file #{file.inspect} not found!")
		end
	end
end
